/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __ENSEMBLE_H
#define __ENSEMBLE_H

#include <list>
#include <vector>
#include <iostream>

namespace Modelib {

///Classe abstraite servant d'interface pour crer des proprits sur les ensembles en creant une classe fille
template<typename T>
struct Propriete
{
  /**operateur virtuelle  surcharger pour crer une propriete
    @param un element gnrique de meme type
           que les elements l'ensemble ou va s'appliquer la propriete
    @return vraie si la proprit est vrifie
    */
  virtual bool operator() (const T &) const= 0;
};

template <typename T >
class Ensemble;

///Classe reprsentant les parties d'ensembles
template <typename T >
class Partie
{
    protected:

    Ensemble<T> * ens;  //!< un pointeur sur l'ensemble
    std::list<unsigned> ids; //!< la liste des elements de la partie

//    template <typename U> friend std::ostream & operator<< (std::ostream & flux, const Partie<U> & );

    /**la classe reprsentant les ensembles*/
    friend class Ensemble<T>;

    T get(unsigned i) const {return *(ens->elements[i]); }

    /**ajoute un indice  la partie*/
    void push(unsigned j)
    {
  	  if( ids.size() && j != ids.back() )
        ids.push_back(j);
      else
        if(!ids.size())
          ids.push_back(j);
    }

    private:
    struct Inter
    {
      bool operator() (unsigned i, unsigned j) const { return i==j; }
    };

    struct Union
    {
      bool operator() (unsigned i, unsigned j) const { return true; }
    };

    template <typename Foncteur>
    Partie parallel(const Partie & p,const Foncteur & f) const
    {
      Partie ret(ens);
      std::list<unsigned>::const_iterator i = ids.begin();
      std::list<unsigned>::const_iterator j = p.ids.begin();
      while( i != ids.end() && j != p.ids.end() )
      {
        if(*i <= *j)
        {
          if( f(*i,*j)==true )
            ret.push(*i);
          ++i;
        }
        else
        {
          if( f(*i,*j)==true )
            ret.push(*j);
          ++j;
        }
      }
      if( i != ids.end() )
        while( i != ids.end()  )
        {
          if( f(*i,*j)==true )
            ret.push(*i);
          ++i;
        }
      else
        while( j != p.ids.end()  )
        {
          if( f(*i,*j)==true )
            ret.push(*j);
          ++j;
        }

      return ret;
    }

    /*-------------*/

    public:

    /**Construction  partir d'un ensemble
     @param ens  un pointeur sur l'ensemble de dpart
     */
    Partie(Ensemble<T> * _ens = 0 ):ens(_ens) {}

    /**constructeur de copie*/
    Partie(const Partie & p) { *this=p;}

    /**operateur d'assignement*/
    Partie & operator= ( const Partie & p)
    {
    	if(this!=&p)
    	{
    		ens = p.ens;
    		ids = p.ids;
    	}
    	return *this;
    }

    /**operateur d'assignement a partir d'un ensemble*/
    Partie & operator= (const Ensemble<T> & _ens)
    {
    	ens = const_cast<Ensemble<T> *>(&_ens);
    	for(unsigned i =0; i < ens->elements.size(); ++i)
    		ids.push_back(i);
    	return *this;
    }

    /**fonction ralisant l'union de cette partie avec une autre
      @param partie  la partie avec laquelle faire l'union
      @return l'union des deux parties*/
    Partie _union(const Partie & partie) const
    {
      return parallel(partie,Union());
    }

    /**fonction ralisant l'intersection de cette partie avec une autre
      @param partie  la partie avec laquelle faire l'intersection
      @return l'intersection des deux parties*/
    Partie _inter(const Partie & partie) const
    {
      return parallel(partie,Inter());
    }

    /**fonction ralisant l'complementaire d'une partie dans cette partie
      @param partie  la partie avec laquelle faire le complementaire
      @return le complementaire de la partie dans cette partie*/
    Partie _compl(const Partie & partie) const
    {
      Partie ret(ens);
      std::list<unsigned>::const_iterator i = ids.begin();
      std::list<unsigned>::const_iterator j = partie.ids.begin();

      while( i!= ids.end() && j != partie.ids.end())
      {
        if(*i <= *j)
        {
          if(*i != *j)
            ret.ids.push_back(*i);
          ++i;
        }
        else
          ++j;
      }
      while(i!=ids.end())
      {
        ret.ids.push_back(*i);
        ++i;
      }

      return ret;
    }

    /**retourne la liste des identifiants de cette partie*/
    inline std::list<unsigned> GetIds() const { return ids;}

   /**retourne la liste des objets de cette partie*/
    std::list<T> Get() const
    {
        std::list<T> ret;
        if(ens)
		    for(std::list<unsigned>::const_iterator i = ids.begin();
                                                  i != ids.end();
                                                  ++i)
			   ret.push_back(ens->Get(*i));
        return ret;
    }

    /**retourne la taille de cette partie*/
    unsigned Size() const { return ids.size();}

    ~Partie() {}
};

///Conteneur reprsentant les ensembles d'elements de type gnrique T
/**
 Ce conteneur est glouton, c'est  dire que l'on ne peut pas lui enlever d'objet \n
 Il faut donc bien le definir a sa contruction
 */
template <typename T >
class Ensemble
{
    private:
    std::vector< T * > elements;
    /**surcharge de flux destine a l'affichage*/
//    template <typename U> 
//	 friend std::ostream & operator<< (std::ostream & flux, const Ensemble<U> & ens);
    /**classe representant une partie de l'ensemble*/
    friend class Partie<T>;
    public:
    typedef Partie<T> Partie;
    /**constructeurr par defaut*/
    Ensemble() {}
    /**constructeur de copie*/
    Ensemble( const Ensemble & ens) { *this=ens;}
    /**operateur d'assignement*/
    Ensemble & operator= ( const Ensemble & ens)
    {
    	if(this!=&ens)
    	{
    	   for(unsigned i = 0; i < elements.size(); ++i)
    	       delete elements[i];
            for(unsigned i2 = 0; i2 < ens.elements.size(); ++i2 )
                elements.push_back( new T(*(ens.elements[i2])) );
        }
    	return *this;
    }

    /**constructeur a partir d'une liste d'elements*/
    Ensemble( const std::list<T>& base)
    {
        typename std::list<T>::const_iterator iterList = base.begin();
    	for( ; iterList != base.end(); ++iterList)
          elements.push_back( new T(*iterList));
    }

    /**recuperation d'une partie verifiant une propriete
      @param propriete  un foncteur drivant de la classe abstraite Propriete
      @return la partie de tout les elements repondant a cette propriete
      */
    Partie partie( const Propriete<T> & p)
    {
        Partie ret(this);
        for(unsigned i=0; i<elements.size(); ++i )
            if( p( *(elements[i]) ))
                ret.ids.push_back( i );
        return ret;
    }

    /**Retourne la partie englobant tous les elements*/
    Partie all()
    {
    	Partie ret;
        ret = *this;
    	return ret;
    }

    /**recupere le i-eme element ajoute a l'ensemble*/
    T Get(unsigned i) const { return *(elements[i]);}
    /**recupere le i-eme element ajoute a l'ensemble*/
    T & Get(unsigned i) { return *(elements[i]);}
    /**renvoie la taille de l'ensemble*/
    unsigned Size() const { return elements.size();}

    /**Ajoute un element a l'ensemble*/
    void Add( const T & element)
    {
        elements.push_back( new T(element) );
    }

    /**Destruvteur*/
    ~Ensemble()
    {
    	for(unsigned i = 0; i < elements.size(); ++i)
    	  delete elements[i];
    }
};

template<typename T>
Partie<T> operator& (const Partie<T>&a, const Partie<T> &b)
{
  return a._inter(b);
}

template<typename T>
Partie<T> operator| (const Partie<T> &a, const Partie<T> &b)
{
  return a._union(b);
}       

template<typename T>
Partie<T> operator/ (const Partie<T> &a, const Partie<T>& b)
{
  return a._compl(b);
}       

template<typename T>
Partie<T> operator/ (const Ensemble<T>& e, const Partie<T> & b)
{
  Partie<T> all;
  all = e;
  return all._compl(b);
}       

template <typename T>
std::ostream & operator<< (std::ostream & flux, const Ensemble<T> & ens)
{
  bool first = true;
  flux << "{";
  for( unsigned i = 0;i < ens.Size();++i)
  {
    flux << (first?"":", ") << ens.Get(i);
    first=false;
  }
  flux << "}";    
  return flux;
}

template <typename T>
std::istream & operator>> (std::istream & flux, Ensemble<T> & ens)
{
  T tmp;
  flux >> tmp;
  ens.Add(tmp);
  return flux;
}

template <typename T>
std::ostream & operator<< (std::ostream & flux,const Partie<T> & p)
{
  bool first = true;
  flux << "{";
  std::list<T> liste = p.Get();
  typename std::list<T>::iterator i;
  for( i = liste.begin(); i != liste.end(); ++i)
  {
      flux << (first?"":", ") << *i;
      first=false;
  }
  flux << "}";
  return flux;
}

}

#endif
