/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __MATRICE_CREUSE_H
#define __MATRICE_CREUSE_H

#include "Config.h"
#include <map>
#include <iostream>
#include <string>
#include <fstream>
#include <typeinfo>
#include <algorithm>
#include <utility>
#include <stdexcept>


namespace Modelib {

///Structure de coordonnes dans la matrice creuse
struct Id {
    unsigned i;
    unsigned j;
    /**Constructeur spcialis*/
    Id(const unsigned _i, const unsigned _j) : i(_i), j(_j) {};
};



///Fonction de comparaison sur la structure Id
template<typename T>
struct Compare
{
  Compare() {};
  /**Foncteur de Comparaison
     @param x
     @param y
     @return vrai si x < y dans l'ordre lxicographique
   */
  bool operator()(const Id& x, const Id& y) const
  { 
      return (x.j < y.j)||(x.j==y.j && x.i <y.i);
  }
};


//!Classe reprsentant les matrices
/**
  Elle est constitue d'un map qui au coordonnes reprsentes par la structure Id fait
  correspondre une value. On ne stocke qu'ainsi ce qui nous interesse: c'est le but
  d'un stockage de matrice creuse.
 */
template<typename T>
class MatriceCreuse
{
protected:

    unsigned nbCols;          //!< pour tester si on dpasse de la matrice ou pas...
    unsigned nbLignes;        //!< pour tester si on dpasse de la matrice ou pas...
    std::map<Id,T,Compare<T> > matrix;    //!< stockage de notre matrice
    T defaultT;                       //!< element par defaut correspondant au 0

    friend class Model;
    
public:
    /**Constructeur par dfaut*/
    MatriceCreuse( T _defaultT = 0 )
      : nbCols(0), nbLignes(0), defaultT( _defaultT )
    {}

    MatriceCreuse(const std::string& nom,T _defaultT = 0);
    bool Open(const std::string& nom);
    //
    T operator() (unsigned i, unsigned j) const;
    T operator() (unsigned i, unsigned j)      ;
    T Get        (unsigned i, unsigned j) const;
    //
    /**
      Seter de coefficient de matrice
      @param i coordonnes colonne dans la matrice
      @param j coordonnes ligne dans la matrice
      @param value value du coefficient
      */
    inline
    void Set     (unsigned i, unsigned j, const T& value)
    {
      const Id S(i,j);
      matrix[S] = value;

      if (i>nbCols)   nbCols  = i;
      if (j>nbLignes) nbLignes= j;
    }
    //
    // En C# : // tout ca pour dire que le C# c beau !
    // public unsigned nbLignes {
    //     get { return nbLignes }
    //     set { nbLignes = value }
    // }
    // L, en C++ on est oblig de faire ces saloperies
    
    /**Geter du nombre de lignes
       @return le nombre de lignes de la matrice creuse
     */
    inline unsigned GetNbLignes() const { return nbLignes;      }
    inline unsigned GetNbLignes()       { return nbLignes;      }
    /**Geter du nombre de colonnes
       @return le nombre de colonnes de la matrice creuse
     */
    inline unsigned GetNbCols()   const { return nbCols;        }
    inline unsigned GetNbCols()         { return nbCols;        }
    /**Obtenir du nombre d'elements
       @return le nombre d' lments de la matrice creuse
     */
    inline unsigned GetSize()     const { return matrix.size(); }
    inline unsigned GetSize()           { return matrix.size(); }

    /**Calculer le taux d'occupation de la matrice. Cette mthode s'effectue 
       en O(n) ou meilleur.
       @return le taux d'occupation de la matrice
    */
    inline double Occupation()
    {
      return static_cast<double>(double(matrix.size()) / double(nbCols * nbLignes));
    }

    /**Sortie graphique
      */
    void Out()
    {
      unsigned i;
      for (unsigned j=1;j<=nbLignes;j++)
      {
        for (i=1;i<=nbCols;i++)
        {
            std::cout << Get(i,j);
            std::cout << ',';
        }
        std::cout << std::endl;
      }
    } ///~

    ~MatriceCreuse()
    {
    }///~
};

#ifdef __no_export
  #include "MatriceCreuse.cxx"
#endif

}

#endif
///~
