/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __NUMARRAY_H_
#define __NUMARRAY_H_

#include "MuteVar.h"
#include "Num.h"

#include <vector>
#include <stdexcept>

namespace Modelib {

//!Classe representant les vecteurs de nombres de base.
/**
   Elle sert surtout pour faire des produits scalaires avec des vecteurs de variables.\n
   Elle est bas sur les vecteurs de la STL et la classe Num. Un vecteur "NumArray" est dclar
   avec un type unique.\n
   Elle possde des operateurs [] qui permettent d'avoir accs  ses composantes.\n\n
   Elle est normalise et  3 hritires IntArray,FloatArray,BoolArray qui servent  la facilit de dclaration.\n
*/


class NumArray
{
  protected:
    std::vector<Num> cstArray; //!< Vecteur de constantes
    MuteVar::VarType type; //!< Type du vecteur
  public:
    /**Constructeur par dfaut et d'initialisation*/
    NumArray(MuteVar::VarType _type = MuteVar::FLOAT, unsigned _size = 0,float _value = 0.0f )
    :cstArray( std::vector<Num>( _size, Num(_type,_value) ) ), type(_type){} //
    /**Constructeur de recopie*/
    NumArray(const NumArray & _n):cstArray(_n.cstArray),type(_n.type) {}
    /**Oprateur de recopie*/
    NumArray & operator= (const NumArray & _n)
    {
      cstArray = _n.cstArray;
      type = _n.type;
      return *this;
    }

    NumArray(const std::vector< Num > & _vect,
             MuteVar::VarType _type = MuteVar::FLOAT)
    :type(_type)
    {
    	if(_vect.size())
    	{
          type = _vect[0].GetType();
          for(unsigned i = 0; i<_vect.size(); ++i)
          {
          	cstArray.push_back( Num(_type, _vect[i].GetValue()));
          }
        }
    }

    NumArray(const std::vector< float > & _vect)
    :type(MuteVar::FLOAT)
    {
    	if(_vect.size())
    	{
          for(unsigned i = 0; i<_vect.size(); ++i)
          {
          	cstArray.push_back( Float(_vect[i]));
          }
        }
    }

    NumArray(const std::vector< double > & _vect)
    :type(MuteVar::FLOAT)
    {
    	if(_vect.size())
    	{
          for(unsigned i = 0; i<_vect.size(); ++i)
          {
          	cstArray.push_back( Float(_vect[i]));
          }
        }
    }
    NumArray(const std::vector< unsigned > & _vect)
    :type(MuteVar::INT)
    {
    	if(_vect.size())
    	{
          for(unsigned i = 0; i<_vect.size(); ++i)
          {
          	cstArray.push_back( Int(_vect[i]));
          }
        }
    }
    
    NumArray(const std::vector< int > & _vect)
    :type(MuteVar::INT)
    {
    	if(_vect.size())
    	{
          for(unsigned i = 0; i<_vect.size(); ++i)
          {
          	cstArray.push_back( Int(_vect[i]));
          }
        }
    }

    /**Oprateur d'indice constant
       @param i l'indice du tableau pour lequel on souhaite rcuprer la value
       @return La value convertie en rel  l'indice i
    */
    float operator[] (unsigned i) const 
    {
    	if( i < cstArray.size())
            return cstArray[i].GetValue();
        else
            throw std::out_of_range("NumArray::operator[] : index out of range");
    }
    /**Oprateur d'indice constant
       @param i l'indice du tableau pour lequel on souhaite rcuprer la value
       @return La constante  l'indice i
    */
    Num & operator[] (unsigned i) 
    {
    	if( i < cstArray.size())
            return cstArray[i];
        else
            throw std::out_of_range("NumArray::operator[] : index out of range");
    }
   /**Geter de taille
       @return La taille du vecteur de constantes
    */
    unsigned Size() const { return cstArray.size();}
    /**Geter de type
       @return Le type du vecteur de constantes
    */
    MuteVar::VarType GetType() const { return type;}
    
    /**operator += : *this = *this + na
        on ajoute respectivement chaque composante des vecteurs
        si leurs tailles correspondent
       @param le vecteur  ajouter
       @return Le type du vecteur de constantes
    */
    NumArray & operator+=( const NumArray & na)
    {
    	if( na.Size() == cstArray.size())
    		for(unsigned i=0; i<cstArray.size();++i)
               cstArray[i].SetValue( cstArray[i].GetValue() + na.cstArray[i].GetValue());
    	else
    		throw std::range_error("NumArray::operator+= : vector size doesn't match");
    	return *this;
    }

    /**operator -= : *this = *this - na
        on ajoute respectivement chaque composante des vecteurs
        si leurs tailles correspondent
       @param le vecteur  soustraire
       @return Le type du vecteur de constantes
    */
    NumArray & operator-=( const NumArray & na)
    {
    	if( na.cstArray.size() == cstArray.size())
    		for(unsigned i=0; i<cstArray.size();++i)
               cstArray[i].SetValue( cstArray[i].GetValue() - na.cstArray[i].GetValue());
    	else
    		throw std::range_error("NumArray::operator+= : vector size doesn't match");
    	return *this;
    }
    
    /**operator *= : *this = *this * f
       @param le rel par lequel on multiplie le vecteur
       @return Le type du vecteur de constantes
    */
    NumArray & operator*=( Num f)
    {
    	for(unsigned i=0; i<cstArray.size();++i)
               cstArray[i].SetValue( cstArray[i].GetValue()*f.GetValue());
        return *this;
    }
    
    ~NumArray() {}

};

///Classe reprsentant les vecteurs de constantes entires.
class IntArray:public NumArray
{
  public:
    IntArray(unsigned _size = 0,int _entier = 0)
    :NumArray(MuteVar::INT, _size, static_cast<float>(_entier)) {}
    IntArray(const std::vector< unsigned > & _vect)
    :NumArray(_vect) {}
    IntArray(const std::vector< int > & _vect)
    :NumArray(_vect) {}

};

///Classe reprsentant les vecteurs de constantes relles.
class FloatArray:public NumArray
{
  public:
    FloatArray(unsigned _size = 0, float flottant = 0.0f)
    :NumArray(MuteVar::FLOAT, _size, flottant) {}
    FloatArray(const std::vector< float > & _vect)
    :NumArray(_vect) {}
    FloatArray(const std::vector< double > & _vect)
    :NumArray(_vect) {}
};

///Classe reprsentant les vecteurs de constantes boolennes.
class BoolArray:public NumArray
{
  public:
    BoolArray(unsigned _size = 0, bool booleen = false)
    :NumArray(MuteVar::BOOL, _size, (booleen?1.0f:0.0f)) {}
    BoolArray(const std::vector< float > & _vect)
    :NumArray(_vect) {}
    BoolArray(const std::vector< double > & _vect)
    :NumArray(_vect) {}
    BoolArray(const std::vector< int > & _vect)
    :NumArray(_vect) {}
    BoolArray(const std::vector< unsigned > & _vect)
    :NumArray(_vect) {}
};

NumArray operator+(const NumArray & na1, const NumArray & na2);

NumArray operator*(const NumArray & na1, Num f);

NumArray operator*( Num f, const NumArray & na1);

NumArray operator-(const NumArray & na1, const NumArray & na2);

}

#endif
