/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __NUMVAR_H_
#define __NUMVAR_H_

#include "Config.h"
#include "Constant.h"
#include "MuteVar.h"
#include <string>

namespace Modelib {

class Model; //vite une dclaration cyclique

//!Classe reprsentant des variables rattaches  un modle
/**
  Le pointeur  t prfr  la rfrence car il premet la dfinition d'une variable neutre NumVar(),
  utile pour reprsenter les constantes dans les expressions par exemple.\n
  La classe est normalise car elle sera utilise avec la STL.\n
  Une variable "NumVar" est ainsi dfinie par constructeur par:
  - un modele
  - une borne infrieure
  - une borne suprieure
  - un type
  - un nom
  Elle peut aussi tre construite par un identifiant et un pointeur sur un modele.\n
  Ce constructeur est utilis surtout pour rcrer des variables de Type NumVar a partir d' un identifiant uniquement. \n
  Elle a trois classes hrites qui premettent juste la simplifiction d'criture
 */

class NumVar
{
  protected:
    VarId varId; //!< identifiant de variable
    Model * modele; //!< modle de dfinition
  public:
    /**Constructeur par dfaut*/
    NumVar():varId(0),modele(0) {}

    NumVar(Model & _modele,
           float _lb= 0,
           float _ub= Infinity,
           MuteVar::VarType _type = MuteVar::FLOAT,
           const std::string & _nom = "unknow" );

    NumVar(Model & _modele,
           MuteVar::VarType _type = MuteVar::FLOAT,
           const std::string & _nom = "unknow" )
    {
    	NumVar(_modele,0,Infinity,MuteVar::FLOAT,_nom);
    }

    NumVar(const Model *,VarId);

    /**Constructeur de recopie*/
    NumVar(const NumVar & _n):varId( _n.varId), modele(_n.modele)  {}

    /**Oprateur de recopie*/
    NumVar & operator= (const NumVar & _n)
    {
      modele = _n.modele;
      varId = _n.varId;
      return *this;
    }

    //Diffrents Geters et Seters
    /**Geter de l'identifiant de variable
       @return L'identifiant de variable sous le format VarId.
     */
    VarId GetVarId() const { return varId;}

    /**Geter du modle
       @return Un pointeur sur le Model o est dfinie la varaible.
     */
    Model * GetModel() const { return modele;}

    float GetValue() const;

    std::string GetName() const;
    void SetName(const std::string &);

    float GetLowerBound() const;
    void SetLowerBound(float _value);
    
    float GetUpperBound() const;
    void SetUpperBound(float _value);
    
    bool IsLowerBoundStrict() const;
    void SetLowerBoundStrict( bool _value = true );
    
    bool IsUpperBoundStrict() const;
    void SetUpperBoundStrict( bool _value = true );
    
    MuteVar::VarType GetType() const;
    void SetType( MuteVar::VarType _type);

    //Destructeur
    ~NumVar() {}
};

/*--------------------------------------------------------------------------------------------*/
/**
  Oprateur d'galit entre deux variables. \n
  @param n1 n2
  @return vrai si elles ont le mme identifiant et si elles sont rattaches au mme modle
*/

inline bool operator==( const NumVar & n1, const NumVar & n2)
{
  return ( n1.GetVarId()==n2.GetVarId() && n1.GetModel()==n2.GetModel() );
}

/*--------------------------------------------------------------------------------------------*/
///Classe rpresentant les variables entires.
class IntVar:public NumVar
{
  public:

#ifdef COMPILE_FOR_PYTHON
    IntVar( Model &_modele,
            float lowerBound = 0,
            float upperBound = Infinity, // ou MaxInt, il faut voir
            const std::string & nom = "unknow" )
    :NumVar(_modele,lowerBound,upperBound,MuteVar::INT,nom)
    {} 	
#else
    IntVar( Model &_modele,
            int lowerBound = 0,
            int upperBound = Infinity, // ou MaxInt, il faut voir
            const std::string & nom = "unknow" )
    :NumVar( _modele,
             (lowerBound > -MaxInt? static_cast<float>(lowerBound) : -Infinity),
             (upperBound <  MaxInt? static_cast<float>(upperBound) :  Infinity),
             MuteVar::INT,nom)
    {}

    IntVar(Model & _modele,
           const std::string & _nom = "unknow" )
    :NumVar(_modele,0,Infinity,MuteVar::INT,_nom)
    {}

#endif	
	
};

///Classe rpresentant les variables relles.
class FloatVar:public NumVar
{
  public:
    FloatVar( Model &_modele,
            float lowerBound = 0,
            float upperBound = Infinity,
            const std::string & nom = "unknow" )
    :NumVar(_modele,lowerBound,upperBound,MuteVar::FLOAT,nom)
    {}
    
    FloatVar(Model & _modele,
           const std::string & _nom = "unknow" )
    :NumVar(_modele,0,Infinity,MuteVar::FLOAT,_nom)
    {} 	
};

///Classe reprsentant les variables boolnnes.
class BoolVar:public NumVar
{
  public:
    BoolVar( Model &_modele,
            const std::string & nom = "unknow" )
    :NumVar(_modele,0.0f,1.0f,MuteVar::BOOL,nom)
    {}
};

#ifdef COMPILE_FOR_PYTHON
	void export_NumVar();
#endif

}

#endif
