/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __FAMILY_H
#define __FAMILY_H

#include "MuteVar.h"
#include "Ensemble.h"
#include "Expr.h"

#include <string>
#include <sstream>
#include <vector>

namespace Modelib {

class Model;

/**
  Tableau de N variables de type T correspondant  un nuplet mathmatique
  */
template<unsigned N,typename T>
struct Nuplet
{
   T tableau[N];

   /**operateur d'aquisition*/
   T & operator[] (unsigned n)
   {
        if(n<N)
            return tableau[n];
        else
            throw domain_error("Nuplet::operator[] : argument is out of range");
    }

   /**operateur d'aquisition constant*/
  T operator[] (unsigned n) const
   {
        if(n<N)
            return tableau[n];
        else
            throw domain_error("Nuplet::operator[] : argument is out of range");
    }

    /**constructeur par dfaut*/
    Nuplet() {}
    
    /**constructeur de copie*/
    Nuplet( const Nuplet & n) { *this = n; }

    /**operateur d'assignement*/
    Nuplet & operator=(const Nuplet &n)
    {
        if( this != &n)
            for(unsigned i=0; i <N;++i)
                tableau[i] = n.tableau[i];
        return *this;
    }
    
    /**destructeur*/
    ~Nuplet() {}

};

/**
  affichage des nuplets
  @param flux  le flux de sortie
  @param nuplet le nuplet  afficher
  @return le flux modifi
  */
template <unsigned N,typename T>
std::ostream & operator<< (std::ostream & flux,const Nuplet<N,T> & nuplet)
{
   flux << "( ";
   for( unsigned i = 0; i < N; ++i)
   {
       flux << nuplet.tableau[i] << " ";
   }
   flux << ")";
   return flux;
}

/**
  lecture de nuplets dans un flux
  @param flux  le flux d'entre
  @param nuplet le nuplet  afficher
  @return le flux lu
  */
template <unsigned N,typename T>
std::istream & operator>> (std::istream & flux, Nuplet<N,T> & nuplet)
{
   for( unsigned i = 0; i < N; ++i)
   {
       flux >> nuplet.tableau[i];
   }
   return flux;
}

/** Foncteur de classification de nuplet pour les graphes*/
template <unsigned I, unsigned N,typename T>
struct Omega: public Propriete< Nuplet<N,T> >
{
  T m;
  /**constructeur spcifiant le sommet rfrence*/
  Omega(T _m):m(_m)  {}
  /**operateur de test*/
  bool operator() (const Nuplet<N,T> & n) const
  {
     return( n.tableau[I] == m );
  }
};

//!Classe reprsentant une famille d'inconnues
/**
  Son utilit rside dans la dclarations de familles de variables.
  Elles ont le mme nom  un indice prs. Elle sont indices sur un ensemble de nuplets.
  Voir le tutoriel pour une meilleure explication.
*/
template <unsigned N, typename T>
class NumVarFamily
{
   std::vector<NumVar> variables;
   typedef Nuplet<N,T> Nuplet_T;
   std::string GetNameNumVar(const std::string & format, const Nuplet_T & nuplet) const
   {
        unsigned j=0;
        std::ostringstream name("");
        for(unsigned i = 0; i < N; ++i)
        {
            while( j < format.length() &&  format[j] != '%' )
            {
                name << format[j];
                ++j;
            }
            if( j < format.length() )
              ++j;
            name << nuplet.tableau[i];
        }
        while( j < format.length() )
        {
           name << format[j];
           ++j;
        }
        return name.str();
   }

  public:

    Ensemble< Nuplet_T > indices;


    /**Constructeur standard
      Usage %->indice ex x%%_%
      @param m  Model o sont defini les variables
      @param lb  la borne inferieure de toutes les varialbles
      @param ub  la borne superieure de toutes les varialbles
      @param type  le type de toutes les variables
      @param ensemble  l'ensemble de nuplet sur lequel les indices sont bass
      @param format  le formar de nom de variable
    */
    NumVarFamily(Model & m,
           float lb,
           float ub,
           MuteVar::VarType type,
           const Ensemble< Nuplet_T > & ensemble,
           const string &format)
    {
    	indices = ensemble;
    	for(unsigned i =0; i < ensemble.Size(); ++i)
    	{
           std::string name =  GetNameNumVar(format,ensemble.Get(i));
           variables.push_back( NumVar(m,lb,ub,type,name) );
    	}
    }

    /**renvoie le vecteur des variables dont l'indices rpond  la proprit
      @param propriete  un foncteur derive de la classe Propriete
      @return un vecteur de variable vrifiant cette propriete
     */
    std::vector<NumVar> GetFamilyPart( const Propriete< Nuplet<N,T> > & propriete)
    {
        return GetFamilyPart( indices.partie(propriete) );
    }

   /**renvoie le vecteur des variables dont l'indices appartiennent  la partie
      @param p  une partie de l'ensemble des indices
      @return  un vecteur de variable dont l'indice appartient  la partie p
     */
   std::vector<NumVar> GetFamilyPart( const typename Ensemble< Nuplet_T >::Partie & p)
    {
    	std::vector<NumVar> retour;
        std::list<unsigned> liste = p.GetIds();
    	for( std::list<unsigned>::iterator iter = liste.begin(); iter != liste.end(); ++iter )
        {
                retour.push_back( variables[*iter] );
        }
       	return retour;
    }
    
    /**rcupre le i-eme element saisi
      @param i  l'ordre de saisie de l'lement
      @return la variable correspondante
      */
    NumVar Get(unsigned i) const { return variables[i];}
    /**rcupre la taille de la famille
      @return le nombre de variables composant la famille */
    unsigned Size() const { return variables.size();}
};

/*--------------------------------------------------------------------------------------------*/
///Classe reprsentant les familles de variables entires.
template <unsigned N, typename T>
class IntVarFamily: public NumVarFamily<N,T>
{
  public:
#ifndef COMPILE_FOR_PYTHON	  
   IntVarFamily( Model &_modele,
              int _lb,
              int _ub,  // ou MaxInt  voir
              const Ensemble< Nuplet<N,T> > & ensemble,
              const std::string & nom )
   :NumVarFamily<N, T>( _modele,static_cast<float>(_lb),static_cast<float>(_ub),MuteVar::INT,ensemble,nom) {}
#else
    IntVarFamily( Model &_modele,
               float _lb,
               float _ub,
               const Ensemble< Nuplet<N,T> > & ensemble,
               const std::string & nom )
    :NumVarFamily<N,T>( _modele,_lb,_ub,MuteVar::INT,ensemble,nom) {}
#endif
};

///Classe reprsentant les familles de variables relles.
template <unsigned N, typename T>
class FloatVarFamily: public NumVarFamily<N,T>
{
  public:
    FloatVarFamily( Model &_modele,
                   float _lb,
                   float _ub,
                   const Ensemble< Nuplet<N,T> > & ensemble,
                   const std::string & nom )
    :NumVarFamily<N,T>( _modele,_lb,_ub,MuteVar::FLOAT,ensemble,nom) {}
};

///Classe reprsentant les vecteurs de variables boolennes.
template <unsigned N, typename T>
class BoolVarFamily: public NumVarFamily<N,T>
{
  public:
    BoolVarFamily( Model &_modele,
                const Ensemble< Nuplet<N,T> > & ensemble,
                const std::string & nom )
    :NumVarFamily<N,T>( _modele,0.0f,1.0f,MuteVar::BOOL,ensemble,nom) {}
};

//Ce qui suit est normalement fortement dconseill en C++
//Nanmoins,  cause d'un problme de typename, typedef est inutilisable
#define Couple Nuplet<2,unsigned>
#define Triplet Nuplet<3,unsigned>

#define CoupleFloatVarFamily  FloatVarFamily<2,unsigned>
#define CoupleIntVarFamily    IntVarFamily<2,unsigned>

#define TripletFloatVarFamily  FloatVarFamily<3,unsigned>
#define TripletIntVarFamily    IntVarFamily<3,unsigned>

#ifdef COMPILE_FOR_PYTHON
	void export_Family();
#endif


}

#endif
