/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __PSFILE_H
#define __PSFILE_H

#include "Model.h"
#include <iostream>
#include <fstream>
#include <string>
#include <stdexcept>

namespace Modelib {

/**
	Classe permettant de gerer/generer des fichier de type PostScript
	
	Les fonctions ont ete repises d'un programme C et converti en C++.
	Il y a eut quelques ajouts et quelques modifications.
*/
class PSFile
{
	std::string   nom;
	std::ofstream psfile;
	
	double hmin, hmax;
	double vmin, vmax;
	double xmin, xmax, xscale;
	double ymin, ymax, yscale;	
	
	protected:
	
	void aux_rectangle(
    double xlo, double xhi,
    double ylo, double yhi,
    double gray,
    char *op
  );
  
	void aux_polygon(
    double x[], double y[],
    int npoints,
    double gray,
    char *op
  );

	void aux_circle(
    double xc, double yc, double radius,
    double gray,
    char *op
  );
  
	void aux_lune(
    double xc, double yc, double radius, double tilt,
    double gray,
    char *op
  );
  
	void define_procs();
	void define_caption_procs();

	void save_scales(
    double xmin, double xmax,
    double ymin, double ymax,
    double hmin, double hmax,
    double vmin, double vmax
  );
  
	void setup_page_state( int xn, int yn);
	void setup_caption_data();
	void aux_end_page();

	public:
		
		PSFile() {};
		PSFile(const std::string& _nom) : nom(_nom) 
		{
			psfile.open(nom.c_str());
			if(!psfile.is_open())
			 throw runtime_error("PSFile::PSFile() : I/O Error");
  		begin_document ();
  		begin_page(1,40,320,0,640, 20,300,750,750, 16 ,1);
		}
		
		~PSFile()
		{
  		end_page();
  		end_document(1);		
			psfile.close();
		}
		
		void close() {psfile.close();};
		
	  /** 
	    Initializes an Encapsulated PostScript file.

	    An Encapsulated PostScript file contains a single figure, with a
	    BoundingBox comment, without an explicit "showpage" command;
	    It is suitable for inclusion in TeX papers and such.  
	
	    This procedure writes the preamble and bounding box, sets
	    coordinate system, clip path, caption font, Postscript ops
	    and constants, etc.
	
	    Client coordinates will range over [xmin __ xmax] x [ymin __ ymax].
	    Plotting area is [hmin __ hmax] x [vmin __ vmax] (in pt).
	    The plotting scales dh/dx and dv/dy must be equal.
	
    	The plotting area is divided implicitly into a grid of /xn/ by /yn/
    	rectangular "cells".
  	*/
		void begin_figure (
    		double xmin, double xmax,
    		double ymin, double ymax,

    		double hmin, double hmax,
    		double vmin, double vmax,

    		int xn, int yn
  	);

 	  /** Finalizes an Encapsulated PostScript file. 
 	  */
		void end_figure ();

  	/** 
    	Initializes a (non-encapsulated, multi-page) PS file.
    	
    	A standard PostScript file contains a document with one or more pages,
    	with explicit "show" commands at the end of each page, and no BoundingBox 
    	comment.  It can be printed by itself.
	
    	The procedure writes the file's preamble, with auxiliary
    	Postscript ops.
	
	    The client must call begin_page and end_page around each page.
	    of the document.
  	*/
		void begin_document ();

  	/**
	    Starts a new page of a PS document. 
	    
	    Sets coordinate system, clip path, dimension constants, etc.
	
	    Client coordinates will range over [xmin __ xmax] x [ymin __ ymax].
	    Plotting area is [hmin __ hmax] x [vmin __ vmax] (in pt).
	    The plotting scales dh/dx and dv/dy must be equal.
	
    	The plotting area is divided implicitly into a grid of /xn/ by /yn/ 
    	rectangular "cells".
  	*/
		void begin_page(
    		int page,
    		double xmin, double xmax,
    		double ymin, double ymax,
		
		    double hmin, double hmax,		
    		double vmin, double vmax,

    		int xn, int yn
  		);

  	/** Finalizes a page: Writes page trailer line, etc. 
  	*/
		void end_page();

  	/** Finalizes a multipage document.  The client must keep track 
        of the number of pages that were written to the file. 
    */
		void end_document (int npages);

  	/** 
    	Starts a new section of a plot. The title is a comment 
  	*/
		void begin_section (char *title);

  	/** 
    	Ends a section of a plot. 
  	*/
		void end_section ();

 	 	/**
    	Sets pen parameters and ink color for line drawing.
    	Dimensions are in *millimeters* 
    */
		void set_pen (
    	double gray,
    	double width,
    	double dashlength,
    	double dashspace
  	);

  	/** 
    	Draws segment from (xa,ya) to (xb,yb) with current pen and gray 
    */
		void draw_segment(
    	double xa, double ya,
    	double xb, double yb
  	);
  	
 		/**
    	Draws a Bezier arc with given control points, using the current pen and gray. 
    */
		void draw_curve(
    	double xa, double ya,
    	double xb, double yb,
    	double xc, double yc,
    	double xd, double yd
  	);
  	
	  /** 
    	Draws the outline of the given rectangle using the current pen. 
    */
		void draw_rectangle(
    	double xlo, double xhi,
    	double ylo, double yhi
  	);

		/** 
			Fills given rectangle with given gray color 
		*/
		void fill_rectangle(
    	double xlo, double xhi,
    	double ylo, double yhi,
    	double gray
  	);

  	/** 
    	Fills rectangle with given gray, then
   		draws its outline with current pen.
  	*/
		void fill_and_draw_rectangle(
    	double xlo, double xhi,
    	double ylo, double yhi,
    	double gray
  	);

  	/**
    	Fills the circle with given center and radius, using the given gray. 
    */
		void fill_circle(
    	double xc, double yc, double radius,
    	double gray
  	);
  	
	  /** 
      Draws the circle with given center and radius, using the current pen and gray. 
    */
		void draw_circle(
        double xc, double yc, double radius
  	);

  	/** 
    	Fills the circle with given center and radius, using the given gray,
    	then draws its outline, using the current pen and gray. 
    */    
		void fill_and_draw_circle(
	    double xc, double yc, double radius,
	    double gray
	  );

  	/** 
    	Fills the lune with given center, radius, and tilt, using the given gray,
    	then draws its outline, using the current pen and gray. 
    */
		void fill_and_draw_lune(
    	double xc, double yc, double radius, double tilt,
    	double gray
  	);
  
  	/**
  	  Fills the polygon (x[1],y[1]),.. (x[n],y[n]) with the given gray level.
  	*/
		void fill_polygon(
    	double x[], double y[],
    	int n,
    	double gray
  	);

  	/**
  	  Draws the contour of the polygon (x[1],y[1]),.. (x[n],y[n]) 
 	   using the current pen and gray.
 	  */
		void draw_polygon(
    	double x[], double y[],
   	 	int n
  	);

  	/**
    	Fills the polygon (x[1],y[1]),.. (x[n],y[n]) with the given gray level,
    	then draws its contour using the current pen and gray.
  	*/
		void fill_and_draw_polygon(
    	double x[], double y[],
    	int n,
    	double gray
  	);

  	/** 
  		Fills triangle /abc/ with given gray level. 
  	*/
		void fill_triangle(
    	double xa, double ya,
    	double xb, double yb,
    	double xc, double yc,
    	double gray
  	);

  	/**
    	Fills the given cell of the current cell grid with the given
    	gray level.
  	*/
		void fill_grid_cell(int xi, int yi, double gray);

  	/** 
    	Draws a reference line perpendicular to the given axis 
    	at the given coordinate value.
  	*/
		void draw_coord_line (char axis, double coord);

  	/** 
    	Draws the cell boundaries with the current pen and gray level. 
    */
		void draw_grid_lines();

  	/**
    	Sets the name and point size of the font to be used by put_label. 
    */
  	void set_label_font(char *font, float size);

  	/**
    	Prints "label" at point (x,y), using the current label font size. 
    	The parameter "xalign" (resp. "yalign)" specifies which point of the string's 
    	bounding box will end up at (x,y): 0.0 means the left (resp. bottom) side,
    	1.0 means the right (resp. top) side.  Default is (0.5, 0.5), meaning 
    	the box will be centered at (x,y). 
    */
		void put_label(
    	char *text, 
    	double x, double y, 
    	float xalign, float yalign
  	);
  	
  	/** 
    	Draws a frame around the plotting area. (The frame 
    	will extend half a line width outside the nominal bounding box.) 
    */
		void draw_frame ();


  	/**
    	Adds a caption text below the drawing, *outside* the nominal
    	bounding box. 
  	*/
		void add_caption(char *txt);


  	/** 
    	Writes a text string to /psfile/, in Postscript form,
    	handling special chars and parentheses.
    
    	Replaces any embedded '\n' by the given /newline/ string.
  	*/
		void put_text(char *text, char *newline);
		
		/**
			Operator overloading
		*/
		PSFile& operator<<(const std::string& str)
		{
			add_caption(const_cast<char*>(str.c_str()));
			return (*this);
		}
		
		PSFile& operator<<(const Model& m);		
};

}

#endif
