/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright
		  notice, this list of conditions and the following disclaimer in the
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __SOLVERAPI_H
#define __SOLVERAPI_H

#include "Solver.h"
#include <string>
#include <vector>
#include <list>
#include <map>

namespace Modelib {

enum SolverName 
{
	CPLEX=0,  //!< utilisation du solveur CPLEX
	GLPK=1,   //!< utilisation du solveur GLPK(GLPSOL)
	GLPSOL=1, //!< idem que GLPK, car le nom de l'executable est glpsol
	NONE=42   //!< si aucune solveur est indiqu
};

typedef SolverName API;

/**
	Classe qui ralise une pseudo-interface entre le 'Model' et
	les diffrents solveurs disponibles

	Cette classe contient toutes les informations relatives
	au modle courant. Elle ne contient pas les informations
	qui sont propres aux diffrents solveurs.
*/
class SolverAPI
{
	SolverName  apiSolver;//!< Le type de solveurs
	std::string path;     //!< Le chemin du soleur
	std::string fileLP;   //!< Le nom du fichier temporaire
	Solver     *solver;   //!< Le solveur courant, celui qui est utilise

	Solution    solution; //!< Conteneur des solutions, permet de garder une ancienne solution complte

  protected:
	//! Crer une instance d'un solveur
	void AllocSolver(const SolverName);

  public:
	SolverAPI(const SolverName = GLPK);
	~SolverAPI();

	//! Initialisation du chemin du solver et du nom du fichier temporaire
	void Init(const std::string& , const std::string& _path="");

	//! Initialisation du chemin du solver et du nom du fichier temporaire et du solveur
	void Init(const SolverName, const std::string&, const std::string& _path="");

	//! Attribuer une nouvelle API
	void SetAPI(const SolverName);
	
	//! Retourne le nom du solveur actuel, ou une erreur si il y en a pas
	inline SolverName GetAPI() const {
		if (!solver)
			return NONE;
		return apiSolver;
	}

	//! Retourne le nom du fichier LP utilis
	inline const char * GetfLP() const {
		return fileLP.c_str();
	}
	//! Attribue le nom du fichier LP utilis
	inline void SetfLP(const std::string& flp) {
		fileLP = flp;
	}

	//
	// On retourne des rfrences car ceci sera utilis dans Model
	// Et SolverAPI existe tant que Model existe !
	//
  public:
	inline std::list<float>      & OutNumVarValues    ()     {return solution.variablesVAL;}
	inline std::list<std::string>& GetNumVarNames     ()     {return solution.variablesNOM;}
	inline std::list<float>      & OutConstraintValues()     {return solution.contraintesVAL;}
	inline std::list<std::string>& GetConstraintNames ()     {return solution.contraintesNOM;}
	inline std::string             & GetObjectiveNames  ()     {return solution.objective.first;}
	inline float                     GetObjectiveValue  ()     {return solution.objective.second;}

	//
	// Interfaces des solveurs
	//
  public:
	inline bool  AddOption(const std::string& opt)               {solver->AddOption(opt); return true;}
	inline void  RemoveOption(const std::string& opt="zzERall")  {solver->RemoveOption(opt);}
	inline float Solve()                                         {return solver->Solve(solution,fileLP);}

	inline void  Infos(std::string& op, std::string& info) {
		return solver->Infos(op,info);
	}

	// Rsultats
	inline bool IsSolved()  const {return solver->Solved(); }
	inline bool IsOptimal() const {return solver->Optimal();}


	//
	// Compatibilits antrieures
	//
  public:
	inline void Options(const std::string& opt) {solver->AddOption(opt);}
	inline void RemoveOptions()  {solver->RemoveOption("zzERall");}

	inline void SetMIP(bool isMIP) {solver->SetMIP(isMIP);}

	//inline std::list<float>      & OutNumVarValues    ()     {return solution.variablesVAL;}
	inline std::list<std::string>& GetNumVarNoms      ()     {return solution.variablesNOM;}
	//inline std::list<float>      & OutConstraintValues()     {return solution.contraintesVAL;}
	inline std::list<std::string>& GetConstraintNoms  ()     {return solution.contraintesNOM;}
};

}

#endif
