/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __LATEXFILE_H
#define __LATEXFILE_H

#include "Model.h"
#include <string>
#include <fstream>

namespace Modelib {

/**
	Cette classe permet de generer un fichier LaTeX.
	Le fichier est considere siimplement comme un flux, tout est
	pris en compte pour que l'affichage HTML soit correct.
*/
class LATEXFile
{
	std::string      nom;	//!< Nom du fichier LaTeX de sortie
	std::string    titre;	//!< Titre dans le fichier LaTeX
	std::ofstream  latex;	//!< Flux sur le fichier

  	bool          beauty;	//!< Je sais plus
	bool       openLater;	//!< Ouvrir plus tard

	// making a fluent LaTeX file with great section/subsection
	bool     isInSection;	//!< Flags qui permettent de savoir dans quel type d'environnement on est
	bool  isInSubsection;	//!< Flags qui permettent de savoir dans quel type d'environnement on est

	protected:

	//! Construit le header du fichier LaTeX
	//! On pourrait simplifier car y'a pas mal de packages demandes
	void BuildHeader()
	{
	  latex << "% Les packages ci-dessous sont tous trouvables sur la ctan" << std::endl;
	  latex << "\\documentclass[11pt,a4paper]{article}" << std::endl;
      latex << "\\usepackage[latin1]{inputenc}" << std::endl;
      latex << "\\usepackage [francais] {babel}" << std::endl;
      latex << "\\usepackage{color}" << std::endl;
      latex << "\\usepackage{listings}" << std::endl;
      latex << "%\\usepackage{fancyhdr}" << std::endl;
      latex << "%\\input{epsf}" << std::endl;
      latex << "\\usepackage{rawfonts}" << std::endl;
      
      latex << "\\setlength{\\hoffset}{-18pt}" << std::endl;
      latex << "\\setlength{\\oddsidemargin}{0pt}     % Marge gauche sur pages impaires" << std::endl;
      latex << "\\setlength{\\evensidemargin}{9pt}    % Marge gauche sur pages paires" << std::endl;
      latex << "\\setlength{\\marginparwidth}{54pt}   % Largeur de note dans la marge" << std::endl;
      latex << "\\setlength{\\textwidth}{481pt}   % Largeur de la zone de texte (17cm)" << std::endl;
      latex << "\\setlength{\\voffset}{-18pt}     % Bon pour DOS" << std::endl;
      latex << "\\setlength{\\marginparsep}{7pt}  % Sparation de la marge" << std::endl;
      latex << "\\setlength{\\topmargin}{10pt}     % Pas de marge en haut" << std::endl;
      latex << "\\setlength{\\headheight}{15pt}   % Haut de page" << std::endl;
      latex << "\\setlength{\\headsep}{10pt}  % Entre le haut de page et le texte" << std::endl;
      latex << "\\setlength{\\footskip}{27pt}     % Bas de page + sparation" << std::endl;
      latex << "\\setlength{\\textheight}{708pt}  % Hauteur de la zone de texte (25cm)" << std::endl;
      latex << "\\newlength{\\larg}" << std::endl;
      latex << "\\setlength{\\larg}{14cm}" << std::endl;
      latex << "\\title{" << titre << "}" << std::endl;
      latex << "\\begin{document}" << std::endl;
      latex << "\\maketitle" << std::endl;
      latex.flush();
   	}

	//! Ecrire la fin du document latex
	void BuildFooter()
	{
		latex << "\\end{document}" << std::endl;;
		latex.flush();
	}

	//! Constructeur par defaut qui ne doit jamais etre appele de l'exterieur
	LATEXFile() {};

	public:

	/**
		Constructeur
		@param _nom			Le nom du fichier LaTeX
		@param _titre		Le titre qui apparaitra ds le rapport LaTeX
		@param _beauty		Je ne sais plus
		@param _openLater	On ouvrira le fichier plutard ?
	*/
	LATEXFile(const std::string& _nom, const std::string& _titre="Programme linaire", bool _beauty=false, bool _openLater=false)
	 : nom(_nom),titre(_titre), beauty(_beauty), openLater(_openLater)
	{
		isInSection    = false;
		isInSubsection = false;

		if (openLater==false)
		{
			latex.open(nom.c_str());
			if( !latex.is_open() )
			 throw runtime_error("LATEXFile::LATEXFile() : I/O Erreur");
			BuildHeader();
		}
	}

	//! Destructeur
	~LATEXFile() { latex.close(); }

	//! Ouvrir notre flux, meme si une instance de LATEXFile a ete definie avant
	void Open()
	{
		latex.open(nom.c_str());
		BuildHeader();
		latex.flush();
	}

	//! Fermer le fichier sans detruire notre instance de HTMLFile
	void Close()
	{
		if (isInSubsection == true) latex << "\t\\end{subsection}" << std::endl;
		if (isInSection    == true) latex << "\\end{section}" << std::endl;
		BuildFooter();
		latex.flush();
		latex.close();
	}

	// Diverses surcharges d'operateur pour faciliter l'ecriture

	LATEXFile& operator<<(const Model& _m);  			//< using <pre> keyword for latex rendering
	LATEXFile& operator<<(const std::string& _s);
	LATEXFile& operator<<(const float _f);
	LATEXFile& operator<<(const char  _c);
	LATEXFile& operator<<(const int   _i);

	//! Ecrire uen section
    void MakeSection(const std::string& _section)
    {
      if (isInSection == false)
      {
        latex << "\\begin{section}{" << _section << "}" << std::endl;
        isInSection = true;
      }
      else
      {
        isInSubsection = true;
        latex << "\t\\begin{subsection}{" << _section << "}" << std::endl;
      }
    }

    //! Fermer une section
    void EndSection()
    {
      if (isInSubsection == true)
      {
        latex << "\t\\end{subsection}" << std::endl;
        isInSubsection = false;
      }
      else
      {
        latex << "\\end{section}" << std::endl;
        isInSection = false; isInSubsection = false;
      }
    }

	//! Faire une nouvelle ligne a la main
	void NewLine()
	{
		latex << std::endl << "\\newline" << std::endl;
		latex.flush();
	}
	
	//! Hack pas beau pour ne pas s'embeter si on veut lier plusieurs flux
    std::ostream & getStream() {return latex;}

};

}

#endif
