/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "Config.h"
#include "Constraint.h"
#include "Model.h"
#include <iostream>

namespace Modelib {

/**Constructeur de l'utilisateur
  @param m  une rfrence sur un Model
  @param id l'identifiant de contrainte
 */
Constraint::Constraint(Model & _m,VarId _id)
{
  model = &_m;
  id = _id;
}

/**Constructeur de l'utilisateur avec un model constant
  @param m  une rfrence constante sur un Model
  @param id l'identifiant de contrainte
 */
Constraint::Constraint(const Model & _m,VarId _id)
{
  model = const_cast<Model *>( &_m );
  id = _id;
}

/**Relaxe une contrainte avec un multiplicateur lagrangien
   ayant une valeur de dpart de lagrangianValue
   si lagrangianValue == -Infinity, le multiplicateur n'est
   pas dfinit
  @param float  la valeur initiale du multiplicateur de Lagrange
 */
bool Constraint::Relax( float lagrangianValue )
{
    //si le modle est dfini on appellera la fonction
    //RelaxConstraint qui effectue la tache
    if(model)
    {
   	    return model->RelaxConstraint( id, lagrangianValue );
    }
    else
        cerr << "Warning in Constraint::Relax : undefined Model" << endl;
    return false;
}

/**Permet de rcuprer le multiplicateur lagrangien
   associ  la contrainte
  @return une instance de LagrangianParam representant ce multiplicateur
 */
LagrangianParam Constraint::GetLagrangianParam() const
{
   return LagrangianParam(const_cast<Model &>(*model),id);
}

/**Permet de renommer une contrainte
  @param name le nouveau de la contrainte
 */
void Constraint::Rename( const std::string & name )
{
    //si le modle est dfini on appellera la fonction
    //RenameConstraint qui effectue la tache
    if(model)
   	    model->RenameConstraint( id, name );
    else
        cerr << "Warning in Constraint::Rename : undefined Model" << endl;
}

/**Permet d'ajouter une Expr au premier membre de la contrainte
   ne permet pas de modifier le second membre
  @param name le nouveau de la contrainte
 */
void Constraint::AddExpr(const Expr & expr)
{
    //si le modle est dfini on appellera la fonction
    //AddToConstraint qui effectue la tache
    if(model)
   	    model->AddToConstraint( id, expr );
    else
        cerr << "Warning in Constraint::AddExpr : undefined Model" << endl;

}

/**Permet de rcuperer une Expr  de la contrainte
 */
Expr Constraint::GetExpr(bool secondMember) const
{
    //si le modle est dfini on appellera la fonction
    //GetRelaxedConstraintExpr qui effectue la tache
    if(model)
   	{
   		Expr retour = model->GetRelaxedConstraintExpr( id, false );
   		if(!secondMember)
           retour -= retour.GetConstant();
   		return retour;
   	}
    else
        cerr << "Warning in Constraint::GetExpr : undefined Model" << endl;
    return Expr();
}

/**Permet de rcuperer une Expr relaxe de la contrainte
 */
Expr Constraint::GetRelaxedExpr(bool secondMember) const
{
    //si le modle est dfini on appellera la fonction
    //GetRelaxedConstraintExpr qui effectue la tache
    if(model)
   	{
   		Expr retour = model->GetRelaxedConstraintExpr( id, true );
   		if(!secondMember)
   		   retour -= retour.GetConstant();
   		return retour;
   	}
    else
        cerr << "Warning in Constraint::GetRelaxedExpr : undefined Model" << endl;
    return Expr();
}

/**Permet de rcuperer le second membrer d'une contrainte
 */
float Constraint::GetSecondMember() const
{
    //si le modle est dfini on appellera la fonction
    //GetSecondMember qui effectue la tache
    if(model)
   	    return model->GetSecondMember( id );
    else
        cerr << "Warning in Constraint::GetSecondMember : undefined Model" << endl;
    return 0.0f;

}

/**Permet de redfinir le second membrer d'une contrainte
 */
void Constraint::SetSecondMember(float value)
{
    //si le modle est dfini on appellera la fonction
    //SetSecondMember qui effectue la tache
    if(model)
   	    model->SetSecondMember( id,value );
    else
        cerr << "Warning in Constraint::GetExpr : undefined Model" << endl;
}

#ifdef COMPILE_FOR_PYTHON
	#include "py_Constraint.cxx"
#endif


}
