/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "Expr.h"
#include "Num.h"
#include "NumArray.h"
#include "NumVar.h"
#include "NumVarArray.h"
#include "Model.h"
#include <map>

namespace Modelib {

using namespace std;

///Fonctions de comparaison sur la classe NumVar qui aide a la simplification
struct Comparaison
{
  Comparaison() {};
  /**Foncteur de Comparaison
     @param x
     @param y
     @return vrai si x < y dans l'ordre des identifiants de variables
   */
  bool operator()(const NumVar& x, const NumVar& y) const
  {
      return (x.GetVarId() < y.GetVarId());
  }
};

/*----------------------------------------------------------------------------*/

// Constructeurs specifique

/**
  Constructeurs  partir d'une seule variable
  @param numVar La variable qui constituera l'expression
  */
Expr::Expr( const NumVar & numVar)
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back(ExprNode( Num(numVar.GetType(), 1.0), numVar));
}

/**
  Constructeurs  partir d'une constante relle
  @param f La constante relle qui constituera l'expression
  */
Expr::Expr( float f )
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back( ExprNode(  Float(f), NumVar() ) );
}

/**
  Constructeurs  partir d'une constante relle
  @param double La constante relle qui constituera l'expression
  */
Expr::Expr( double f )
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back( ExprNode(  Float((float)f), NumVar() ) );
}

/**
  Constructeurs  partir d'une constante entire
  @param i La constante entire qui constituera l'expression
  */
Expr::Expr( int i )
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back( ExprNode(  Int(i), NumVar() ) );
}

/**
  Constructeurs  partir d'une constante entire non signe
  @param i La constante entire qui constituera l'expression
  */
Expr::Expr( unsigned i )
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back( ExprNode(  Int(i), NumVar() ) );
}

/**
  Constructeurs  partir d'une constante
  @param num La constante qui constituera l'expression
  */
Expr::Expr( const Num & num)
:isSimplified(false),isParametric(false),sansConstantes(false)
{
  nodeList.push_back( ExprNode(  num, NumVar() ) );
}

/**
  Constructeurs  partir d'une constante et d'une variable \n
  Le produit constante qui constituera l'expression \n
  @param num  La constante
  @param numVar  La variable
  */
Expr::Expr( const Num & num, const NumVar & numVar)
:isSimplified(false),isParametric(false),sansConstantes(false)
{
    nodeList.push_back(ExprNode( num, numVar));
} 

/**
  Constructeurs  partir d'une constante et de deux variable \n
  Le produit constante qui constituera l'expression \n
  @param num  La constante
  @param numVar  La variable
  @param numVar  La variable optionnelle
  */
Expr::Expr( const Num & num, const NumVar & numVar, const LagrangianParam & mult)
:isSimplified(false),isParametric(true),sansConstantes(false)
{
    nodeList.push_back(ExprNode(num,numVar,mult));
}


/*----------------------------------------------------------------------------*/

/**ASS: ajout sans simplification
  @param num
  @param numVar
  @return L'Expr {expr=expr+num*numVar} non simplifie
  */
Expr & Expr::ASS( const Num & num, const NumVar & numVar )
{
    nodeList.push_back( ExprNode( num, numVar) );
    isSimplified = false;
    return *this;
}

/**ASS: ajout sans simplification
  @param num
  @param numVar
  @return L'Expr {expr=expr+num*numVar} non simplifie
  */
Expr & Expr::ASS( const Num & num, const NumVar & numVar, const LagrangianParam & mult )
{
    nodeList.push_back( ExprNode( num, numVar, mult ) );
    isSimplified = false;
    isParametric = true;
    return *this;
}

/*----------------------------------------------------------------------------*/

/**
 Additionne une expression avec l'expression this
 @param _expr
 @return L'Expr {expr=expr+_expr} non simplifie
 */
Expr & Expr::operator+=( const Expr & _expr)
{
  list<ExprNode>::const_iterator src = _expr.nodeList.begin();

  for(;src!=_expr.nodeList.end();++src)
  {
    nodeList.push_back( *src );
  }
  isSimplified = false;

  return *this;
}

/*----------------------------------------------------------------------------*/

/**
  Soustrait une expression avec l'expression this
  @param _expr
  @return L'Expr {expr=expr-_expr}  non isSimplified
 */
Expr & Expr::operator-=( const Expr & _expr)
{
  list<ExprNode>::const_iterator src = _expr.nodeList.begin();

  for(;src!=_expr.nodeList.end();++src)
  {
    nodeList.push_back( ExprNode( Num(-src->GetCoefficient().GetValue()),
                                  src->GetVariable(),
                                  src->GetLagrangianParam()  ) );
  }
  isSimplified = false;

  return *this;
}


/*----------------------------------------------------------------------------*/

/**
 Multiplie une expression par une constante
 @param a
 @return L'Expr {expr=a*this} non simplifie
 */
Expr & Expr::operator*=( const Num & a)
{
  list<ExprNode>::iterator src = nodeList.begin();

  for(;src!=nodeList.end();++src)
  {
    src->SetCoefficient( a.GetValue() * src->GetCoefficient().GetValue() );
  }
  isSimplified = false;

  return *this;
}

/*----------------------------------------------------------------------------*/

/**
  Fonction Simplify met l'expr sous la forme a1*x1+...+an*xn tel que les xi
  n'apparaissent qu'une fois si ai est non nul.
  @param avecLagrangianParam permet de tenir compte des multiplicateurs de Lagrange
  @param sansConstante permet de ne pas tenir compte des constantes
 */

void Expr::Simplify( bool avecLagrangianParam, bool sansConstante  )
{
  // si l'expression est dj simplifie et que l'on travaille sans les lagrangiens
  if(isSimplified && !avecLagrangianParam)
      return;
      //on quitte

    map<NumVar,float,Comparaison> tmp;
    //table temporaire permettant d'associer un coefficient pour chaque variable
    map<NumVar,float,Comparaison>::iterator trouve;
    //iterateur sur cette table

    list<ExprNode>::iterator pos = nodeList.begin();
    //iterateur de liste de bloc coefficient*Variable initialis au dbut de la liste

    list<ExprNode> tmpList;
    //liste temporaire de bloc coefficient*Variable

    while( pos != nodeList.end() )
    {
      float coef = pos->GetCoefficient().GetValue();
      NumVar var = pos->GetVariable();
      LagrangianParam mult = pos->GetLagrangianParam();

      if( avecLagrangianParam )
      {
          if(mult.GetModel())
             coef = coef*mult.GetValue();
          trouve = tmp.find(var);
          if(trouve != tmp.end())
             trouve->second+=coef;
          else
             tmp[var] = coef;
      }
      else
      {
        if(mult.GetModel() == 0)
        {
            trouve = tmp.find(var);
            if(trouve != tmp.end())
               trouve->second+=coef;
            else
               tmp[var] = coef;
        }
      }
      pos = nodeList.erase(pos);
    }
  
    for(trouve = tmp.begin(); trouve != tmp.end(); ++trouve)
    {
      float coef = trouve->second;
      NumVar var = trouve->first;
      if( coef  != 0.0f )
      {
        if( var.GetModel() )
        {
        	  MuteVar::VarType type = var.GetType();
            tmpList.push_back(ExprNode(Num(type,coef),var));
        }
        else
           if( !sansConstante)
             tmpList.push_front(ExprNode(Num(MuteVar::FLOAT,coef),var));
      }
    }
    nodeList.splice(nodeList.begin(),tmpList);

  isSimplified = true;

  return;
}


/**
   Permet de tester si l'expression est constitue seulement d'une constante
   Fonction utilis dans pour dans Constrain::CheckIfIsBound()
   @return vrai si l'expression se reduit  une constante
*/
bool Expr::IsNum()
{
  return ( Size()==1 && (nodeList.front().GetVariable().GetModel() == 0) );
}


/**
   Permet de tester si l'expression est constitue seulement d'une variable
   Fonction utilis dans pour dans Constrain::CheckIfIsBound()
   @return vrai si l'expression se reduit  une variable
*/

bool Expr::IsVar()
{
   return ( Size()==1 && (nodeList.front().GetVariable().GetModel() != 0)
                      && nodeList.front().GetCoefficient().GetValue() == 1.0f);
}

/**
   Permet de rcuperer la partie constante de l'expression
   @param avecLagrangianParam permet de tenir compte de la prsence du multiplicateur de Lagrange
   @return la constante de l'expression
*/
float Expr::GetConstant(bool avecLagrangianParam ) const
{
  list<ExprNode>::const_iterator src = nodeList.begin();
  float retour=0;
  while( !( src==nodeList.end() || ( isSimplified && src->GetVariable().GetModel()!=0) ) )
  {
    if( avecLagrangianParam && src->GetVariable().GetModel()==0 )
      if( src->GetLagrangianParam().GetModel() )
        retour += src->GetCoefficient().GetValue()*src->GetLagrangianParam().GetValue();
      else
        retour += src->GetCoefficient().GetValue();
    else
      if(!src->GetLagrangianParam().GetModel() && src->GetVariable().GetModel()==0 )
        retour += src->GetCoefficient().GetValue();
    ++src;
  }
  return retour;
}

/**
   Permet de rcuperer la valeur de l'expression
   @param avecLagrangianParam permet de tenir compte de la prsence du multiplicateur de Lagrange
   @return la valeur de l'expression
*/
float Expr::Evalue( bool avecLagrangianParam ) const
{
  float retour = 0.0f;
  list<ExprNode>::const_iterator src = nodeList.begin();
  
  for( ; src != nodeList.end(); ++src)
  {
    if( avecLagrangianParam )
    {
    	if( src->GetVariable().GetModel() )
      {
      	if( src->GetLagrangianParam().GetModel() )
          retour += src->GetCoefficient().GetValue()
                    *src->GetLagrangianParam().GetValue()
                    *src->GetVariable().GetValue();
        else
          retour += src->GetCoefficient().GetValue()
                    *src->GetVariable().GetValue();
      }
      else
      {
      	if( src->GetLagrangianParam().GetModel() )
          retour += src->GetCoefficient().GetValue()
                    *src->GetLagrangianParam().GetValue();
        else
          retour += src->GetCoefficient().GetValue();
      }
    }
    else
    {
    	if( src->GetVariable().GetModel() )
          retour += src->GetCoefficient().GetValue()
                    *src->GetVariable().GetValue();
      else
          retour += src->GetCoefficient().GetValue();
    }
  }
  return retour;
}

/*----------------------------------------------------------------------------*/

/**
   Operateur d'addition d'expression
   @param expr1
   @param expr2
   @return expr1 + expr2
*/
Expr operator+( const Expr & expr1, const Expr & expr2)
{
  Expr retour = expr1;
  retour+=expr2;
  return retour;
}


/**
   Operateur de soustraction d'expression
   @param expr1
   @param expr2
   @return expr1 - expr2
*/
Expr operator-( const Expr & expr1, const Expr & expr2)
{
  Expr retour = expr1;
  retour-=expr2;
  return retour;
}


/**
   Operateur de multiplication d'expression
   @param expr une expression
   @param a une constante
   @return a*expr
*/
Expr operator*( const Expr & expr, const Num & a)
{
  Expr retour = expr;
  retour*=a;
  return retour;
}


/**
   Operateur de multiplication d'expression
   @param a une constante
   @param expr une expression
   @return a*expr
*/
Expr operator*( const Num & a , const Expr & expr)
{
  Expr retour = expr;
  retour*=a;
  return retour;
}


/**
  Operateur Num * NumVar crer une expression  partir d'une constante et d'une variable
  @param num
  @param numVar
  @return L'Expression {+num*numVar}
*/
Expr operator*( const Num & num, const NumVar & numVar)
{
  return Expr(num,numVar);
}


/**
  Operateur NumVar * Num crer une expression  partir d'une constante et d'une variable
  @param numVar
  @param num
  @return L'Expression {+num*numVar}
*/
Expr operator*( const NumVar & numVar, const Num & num)
{
  return Expr(num,numVar);
}


/**
  Operateur NumVarArray * NumArray produit scalaire qui va  crer une expression \n
   partir d'un vecteur de constantes et d'un vecteur de variables
  @param nva
  @param na
  @return L'Expr na[0]*nva[0] + na[1]*nva[1] + ... + na[n-1]*nva[n-1]
*/
Expr operator*( const NumVarArray & nva, const NumArray & na)
{
  Expr retour;
  unsigned minSize = nva.Size() < na.Size() ? nva.Size() : na.Size();
  for(unsigned i = 0; i < minSize; ++i)
    retour+=(nva[i])*(na[i]);
  return retour;
}


/**
  Operateur NumArray * NumVarArray  produit scalaire qui va  crer une expression \n
   partir d'un vecteur de constantes et d'un vecteur de variables
  @param na
  @param nva
  @return L'Expr na[0]*nva[0] + na[1]*nva[1] + ... + na[n-1]*nva[n-1]
*/
Expr operator*( const NumArray & numArray , const NumVarArray & numVarArray)
{
  return numVarArray * numArray;
}

/*----------------------------------------------------------------------------*/

/**
  Operateur de surcharge de flux <<
  On simplifie la fonction avant affichage
  @param flux Flux  surcharger
  @param expr L'Expr  afficher
  @return le flux surcharg
*/

std::ostream & operator<< (std::ostream & flux, const Expr & expr )
{
  if(!expr.isSimplified)  //si l'expression n'est pas simplifie
  {
   Expr e = expr; //on la duplique
   e.Simplify(expr.isParametric,expr.sansConstantes);  //On la simplifie
   flux << e;     //et on l'affiche avec le meme operateur
   return flux;
  }
  else
  {
    list<ExprNode>::const_iterator i = expr.nodeList.begin();

    if(i!=expr.nodeList.end())   //Pour les expressions vides
    {
      bool affSigne = true;  //sert pour ne pas afficher le premier signe +
      for( ;i!= expr.nodeList.end(); ++i) //On parcourt la liste de couple Num*NumVar
      {
        float coef = i->GetCoefficient().GetValue(); //On recupere coefficient
        string name;
        if(i->GetVariable().GetModel())
              name = i->GetVariable().GetName(); //On recupere le nom

        if( name != defaultGetName)
        {
          if(affSigne) //si on est au dbut de l'expression, on n'affiche pas de '+'
          {
              //si la variable est dfinie et le coef diffrent de 1
              if( coef != 1.0f &&  coef != -1.0f && name != noModel)
                flux << coef << name;
              else
                if( name != noModel )
                  flux << (coef == 1.0f?"":"-") << name;
                else
                  flux << coef;
              affSigne = !affSigne;
          }
          else
          //idem mais avec l'affichage du signe.
          if( coef != 1.0f &&  coef != -1.0f && name != noModel )
             flux << (coef > 0?"+":"") << coef << name;
          else
             if( name != noModel )
               flux << (coef == 1.0f?"+":"-") << name;
             else
               if( coef != 0.0 )
                 flux << '+' << coef;
        }
      }
    }
    return flux;
  }
}


/*----------------------------------------------------------------------------*/

/**
  Sum additionne les termes d'un vecteur de variables
  @param _x Le vecteur de variables  sommer
  @return L'Expr _x[0] + _x[1] + ... + _x[n-1]
*/

Expr Sum(const NumVarArray & _x)
{
  Expr retour;
  for(unsigned i=0;i < _x.Size();++i)
     retour+=Expr(_x[i]);
  return retour;
}


}
