/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//:~ sof
#include "Config.h"
#include "FichierMPS.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <string>
#include <algorithm>
#include <cctype>
#include <map>
#include <utility>
#include <vector>

namespace Modelib {

using namespace std;

FichierMPS::FichierMPS()
{
  objectifName="";
  isMax=false;
  integerMarker=false;
}

bool FichierMPS::Open(const string& nomFile)
{
  ifstream ifile(nomFile.c_str());

  if (!ifile.is_open())
     return false;
     
  int section = -1;
  string s;
  
  map<string,int> sectionMap;
  map<string,int>::iterator sectionMapIter;

  sectionMap["NAME"] = 0;
  sectionMap["ROWS"] = 1;
  sectionMap["COLUMNS"] = 2;
  sectionMap["RHS"] = 3;
  sectionMap["BOUNDS"] = 4;
  sectionMap["ENDATA"] = 5;

  while (getline(ifile,s,'\n'))
  {
    istringstream ligne(s);
    string first;

    ligne >> first;
    
    sectionMapIter = sectionMap.find( first );
    if( sectionMapIter != sectionMap.end() && s[0] != ' ' )
    {
    	if( section < sectionMapIter->second  )
    	{
    		section = sectionMapIter->second;
    	}
    	if( section == 0 )
    	{
      	  ligne >> nom;
    		  commentaires.push_back(pair<long,string>(0,"\\" + nom));
    	}
    }
    else
    {
    	switch(section)
    	{
    		case 0:
    		break;
    		case 1:
    		  ParseROWS(ligne.str());
    		break;
    		case 2:
    		  ParseCOLUMNS(ligne.str());
    		break;
    		case 3:
     		  ParseRHS(ligne.str());
   		  break;
    		case 4:
    		  ParseBOUNDS(ligne.str());
    		break;
    		case 5:
    		break;
    		default:
          return false;
    		break;
    	}
    }
  }
  nom = objectifName;
  return true;
}

void FichierMPS::ParseROWS(const std::string& row)
{
    istringstream s(row);

    string type;
    string name;
    unsigned comp=99;

    s >> type >> name;

    if( type == "N" )
    {
    	if( objectifName == "" )
    	{
    		objectifName = name;
    	}
    }
    else
    {
      if( type == "G" )
      {
      	comp=2;
      }
      if( type == "L" )
      {
      	comp=0;
      }
      if( type == "E" )
      {
      	comp=5;
      }

      std::map< std::string, plGlobalNode * >::iterator constraintMapIter = constraintMap.find(name);
      if( constraintMapIter == constraintMap.end() )
      {
      	plGlobalNode * newCons = new plGlobalNode(name,list<plExprNode *>(), plSimpleConstraint("",comp));
        exprConstraint.push_back( newCons );
        constraintMap[name]=newCons;
      }
    }
}

void FichierMPS::ParseCOLUMNS(const std::string& col)
{
    istringstream s(col);
    string variable;
    string constraint;
    string coefficient;


    s >> variable;

    while( s >> constraint >> coefficient )
    {
      if( constraint == "'MARKER'" )
      {
        if( coefficient == "'INTORG'")
          integerMarker = true;
        else
          integerMarker = false;
        return;
      }
      else
      {
        if( constraint == objectifName )
      	{
           exprObj.push_back( new plExprNode(variable, coefficient, PLUS ) );
        }
      	else
      	{
          std::map< std::string, plGlobalNode * >::iterator constraintMapIter = constraintMap.find(constraint);
          if( constraintMapIter != constraintMap.end() )
          {
            constraintMapIter->second->expr.push_back( new plExprNode(variable, coefficient, PLUS ) );
          }
      	}
      }
    }

    if( integerMarker )
    {
    	exprGeneral.push_back( variable );
    }
}

void FichierMPS::ParseRHS(const std::string& rhs)
{
    istringstream s(rhs);
    string constraint;
    string coefficient;
    string rhsName;

    s >> rhsName;
    while( s >> constraint >> coefficient )
    {
        std::map< std::string, plGlobalNode * >::iterator constraintMapIter = constraintMap.find(constraint);
        if( constraintMapIter != constraintMap.end() )
        {
          constraintMapIter->second->cont.bound = coefficient;
        }
    }
}

void FichierMPS::ParseBOUNDS(const std::string& bnd )
{
    istringstream s(bnd);
    
    string variable;
    string coefficient = "-Inf";
    string bndName;
    string bndType;

    s >> bndType >> bndName >> variable >> coefficient;

    if( bndType == "BV" )
    {
    	exprBinary.push_back(variable);
      return;
    }

    plConstraintNode * tmp;
    if( bndType == "MI" )
    {
      tmp = new plConstraintNode("1",variable,"-inf",LOWER_THAN );
    }
    
    if( bndType == "LO" || bndType == "LI")
    {
      tmp = new plConstraintNode("1",variable,coefficient,LOWER_THAN );
    }
    
    if( bndType == "UP" || bndType == "UI" )
    {
      tmp = new plConstraintNode(variable,"1",coefficient,"inf",2,99);
    }
    
    if( bndType == "FX")
    {
      tmp = new plConstraintNode("1",variable,EQUAL,coefficient);
    }

    if( bndType == "FR")
    {
      tmp = new plConstraintNode (variable,"1","-inf","inf",LOWER_THAN,LOWER_THAN);
    }
    
    exprBounds.push_back( tmp );
}

}
