/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following
	conditions are met:

		* Redistributions of source code must retain the above copyright
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "GlpkSolver.h"
#include "Utils.h"
#include "System.h"

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>

namespace Modelib {

using namespace std;

/**
	Constructeur
*/
GlpkSolver::GlpkSolver()
{
	nomSolver = "glpsol";
	options.push_back("--cpxlp"); // on dit a GLPSOL d'utiliser le format LP
}

/**
	Destructeur
*/
GlpkSolver::~GlpkSolver()
{
	//
}


float GlpkSolver::Solve(Solution& sol, const string& nameFile)
{
	// effacer l'ancienne solution
	sol.Clear();
	tempName = ".tmp." + nameFile;
	string solName(nameFile);
	Utils::Replace(solName, ".lp", ".sol");

	// Creer le fichier temporaire
	try {
		CreerFichierTemporaire(nameFile);
	}
	catch(exception& e)	{
		cout << e.what() << endl;
	}

	// Appeler le solveur
	Resoudre(solName);
	
	try {
	// Interpreter les rsultats
	Interprete(solName, sol);
	}
	catch (exception& e)
	{
		cout << e.what() << endl;
	}

	return sol.objective.second;
}


/**
	Ecrire dans le fichier temporaire au maximum 255 caractres par ligne
*/
ostream& GlpkSolver::EcrireBonneLargeur(ostream& out, const string& str)
{
	if (str.length() < 255)
		out << str << endl;
	else
	{
		string temp;
		unsigned nb_chars=0;
		for (string::const_iterator iter=str.begin();iter!=str.end();++iter,nb_chars++)
		{
			temp.push_back(*iter);
			if(*iter == '+' || *iter == '-')
			{
				if(nb_chars + temp.length() < 255)
				{
					out << temp;
					temp.clear();
				}
				else
				{
					nb_chars = 0;
					out << endl << temp ;
					temp.clear();
				}
			}
		}
		out << temp << endl;
	}
	return out;
}


/**
	Crer une fichier temporaire pour GLPK
	Passer outre la limitation de la lecture des fichiers ayant
	que 256 caractres de largeur (utilisation de la fonction
	'EcrireBonneLargeur').
*/
void GlpkSolver::CreerFichierTemporaire(const string& nameFile)
{
	ofstream mod(tempName.c_str());
	if(!mod.is_open())
		throw runtime_error("(GlpkSolver::CreerFichierTemporaire) - I/O Error pour " + tempName);

	ifstream in(nameFile.c_str());
	if(!in.is_open())
		throw runtime_error("(GlpkSolver::CreerFichierTemporaire) - I/O Error");
		
	string s;
	while (getline(in, s))
	{
		EcrireBonneLargeur(mod, s);
	}
}


/**
	Appeler le solveur et lui spcifier de mettre le rsultat dans
	solName
*/
void GlpkSolver::Resoudre(const string& solName)
{
	string cmd = nomSolver;
	// ajouter les options
	for (vector<string>::const_iterator iter=options.begin(); iter!=options.end(); ++iter) {
		cmd += " ";
		cmd += (*iter);
	}
	// ajouter le nom du fichier LP
	cmd += " ";
	cmd += tempName;

	// avoir la sortie fichier
	cmd += " -o ";
	cmd += solName;
	
	// Appel systme
	System::call(cmd.c_str());
}


/**
	Interpreter la solution retourne
	
	@param solName Le nom du fichier contenant la sortie de GLPSOL
	@param sol     Le conteneur de la solution
*/
void GlpkSolver::Interprete(const string& solName, Solution& sol)
{
	ifstream inFile(solName.c_str());
	if(!inFile.is_open())
	    throw runtime_error("(GlpkSolver::Inteprete) - I/O Error");

	string s;
	bool Debut = true;
	bool Contraintes=false,Variables=false;


	cout << "Interpreter le fichier " << solName << endl;


	while (getline(inFile,s))
	{
		// On rcupre la value de la fonction objectif
		if (Debut)
		{
			isSolved = true;
			string::size_type place;
			if ((place=s.find("Status:")) != string::npos)
			{
				string reste,tmp;
				istringstream sstr(s);
				sstr >> tmp >> reste;
				tmp.clear();
				while(sstr)
				{
					sstr >> tmp;
					reste.push_back(' ');
					reste += (tmp);
					reste.push_back(' ');
					tmp.clear();
				}

				typeRetour = reste;
				if ((place=typeRetour.find("OPTIMAL")) != string::npos)
					isOptimal = true;
			}
			else  if ((place=s.find("Objective:")) != string::npos)
			{
				string _nom;
				string reste[10];
				float _val;
				string c;
				unsigned i=1;

				//cout << "Ligne de solution: " << s << endl;
				istringstream sstr(s);
				sstr >> reste[0] >> _nom >> c >> _val;
				while(sstr)
					sstr >> reste[i++];
				sol.objective = make_pair(_nom, _val);

				// travail sur (MINimum) etc. pour avor un mot sympa
				if ((place=reste[1].find("MIN")) != string::npos)
					typeOperation = "Minimisation";
				else
					typeOperation = "Maximisation";
				Debut = false;
			}
			else
				continue;
		
		}
		else if (s.find("Row ") != string::npos) // le p
		{
			// on accelere le passage des ------------- ------------- --------------
			getline(inFile,s);
			Contraintes = true;
		}
		else if (s.find("Column ") != string::npos) // le p
		{
			Contraintes = false;
			// on accelere le passage des ------------- ------------- --------------
			getline(inFile,s);
			Variables = true;
		 }
		else if (Contraintes)
		{
			//    No.   Row name   St   Activity     Lower bound   Upper bound    Marginal
			istringstream sstr(s);
			unsigned id;
			string nom;
			string st;
			float var;
			string lB, uB, M;
			sstr >> id >> nom >> st >> var >> lB >> uB >> M;

			if (nom == "")
				Contraintes = false;

			sol.contraintesVAL.push_back(var);
			sol.contraintesNOM.push_back(nom);
		}
		else if (Variables)
		{
			//    No.   Row name   St   Activity     Lower bound   Upper bound    Marginal
			istringstream sstr(s);
			unsigned id;
			string nom;
			string st;
			string varstr;
			string lB, uB, M;
			sstr >> id >> nom >> st >> varstr >> lB >> uB >> M;

			if (nom == "")
				Variables = false;

			if (st != "*"
		    &&  st != "NL"
			&&  st != "B"
			&&  st != "NS"
			&&  st != "NU")
			{
			   varstr = st;
			}

			istringstream tmp(varstr);
			float var;
			tmp >> var;
			sol.variablesVAL.push_back(var);
			sol.variablesNOM.push_back(nom);
		}
	}
	inFile.close();
}


}

