/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "NumVar.h"
#include "Constant.h"
#include "Model.h"
#include <iostream>

namespace Modelib {

using namespace std;


/*-----------------------------------------------------------------------------------------------------------*/
/**!
   Constructeur requierant un modele pour tre utiliser. Facilite la lisibilit des dclarations de variables
   et permet de s'assurer que le modele existe et est valide.\n
   Fait appel  la fonction AddVar de la classe Model qui est protge au sein de Model. Ceci est possible
   car la classe NumVar est dclare friend dans la classe Model. Ceci permet un controle de la dclaration
   de variable impossible si AddVar tait publique.\n
   Un test permet de detecter l'existence du modele mais pas sa validit.
   @param _Model modle d'existence de la variable
   @param _lb Value de borne infrieure
   @param _ub Value de borne suprieure
   @param _type Type de la variable
   @param _nom nom de la variable
   @see Model::AddVar
  */

NumVar::NumVar(Model & _modele,
       float _lb,
       float _ub,
       MuteVar::VarType _type,
       const std::string & _nom )
:varId( _modele.AddVar( MuteVar( _type, 0, _lb, _ub),_nom ) ), modele(&_modele)
{}

/**
   Constructeur permettant de recrer une variable a partir d'un identifiant et d'un pointeur constant sur un modele.\n
   Si le modele n'existe pas ou si l'identifiant ne represente rien alors la variable est neutre.\n
   @see Model::GetName
 */
NumVar::NumVar( const Model * _modele,VarId _varId)
:varId(0), modele(0)
{
   //si le modle existe et si la variable est definie
   if( _modele && _modele->GetName( _varId ) != defaultGetName )
   {
      modele = const_cast<Model *>(_modele);
      //le modele est constant il faut donc le transtyper
   	  varId = _varId;
   }
   else
     if(_modele)
       cerr << "Warning in NumVar::NumVar : undefined variable" << endl;
     else
       cerr << "Warning in NumVar::NumVar : undefined model" << endl;

}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter de la value de la variable represente dans le modele par une MuteVar.
  @return La value de varaible aprs rsolution du problme
  @see  Model::GetMuteVar
 */
float NumVar::GetValue() const
{
   //si le modele existe
   if(modele)
     return modele->GetMuteVar(varId).GetValue();
     //on recupere l'instance de MuteVar puis sa valeur
   else
   {
     cerr << "Warning in NumVar::GetValue : undefined model" << endl;
   	 return 0;
     //on retourne 0.
   }
}

/**
  Geter du nom associ  la variable represent dans le modele par un std::string.
  @return La nom associ  la varaible
  @see  Model::GetName
 */
std::string NumVar::GetName() const
{
   //si le modele existe
	if(modele)
    return modele->GetName(varId);
     //on recupere le nom de la variable dans Model::unkownMap
  else
    cerr << "Warning in NumVar::GetName : undefined model" << endl;
  return noModel;
}

/**
  Seter du nom associ  la variable represent dans le modele par un std::string.
  @param La nom associ  la varaible
  @see  Model::SetName
 */
void NumVar::SetName(const std::string & name)
{
  //si le modele existe
	if(modele)
     modele->SetName(varId, name );
     //on redefinit le nom de la variable dans Model::unkownMap
  else
    cerr << "Warning in NumVar::GetName : undefined model" << endl;
}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter de la borne infrieure de la variable represente dans le modele par une MuteVar.
  @return float La value de la borne infrieure
  @see Model::GetMuteVarLB
 */
float NumVar::GetLowerBound() const
{
	if(!modele)
    cerr << "Warning in NumVar::GetLowerBound : undefined model" << endl;
  return (modele?modele->GetMuteVarLB(varId):0);
}

/**
  Seter de la borne infrieure de la variable represente dans le modele par une MuteVar.
  @param _value La value de la borne infrieure
  @see Model::SetMuteVarLB
 */
void NumVar::SetLowerBound(float _value)
{
	if(!modele)
    cerr << "Warning in NumVar::SetLowerBound : undefined model" << endl;
	if(modele) modele->SetMuteVarLB(varId,_value);
}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter de la borne suprieure de la variable represente dans le modele par une MuteVar.
  @return float La value de la borne suprieure
  @see Model::GetMuteVarUB
 */
float NumVar::GetUpperBound() const
{
	if(!modele)
    cerr << "Warning in NumVar::GetUpperBound : undefined model" << endl;
	return (modele?modele->GetMuteVarUB(varId):0);
}

/**
  Seter de la borne suprieure de la variable represente dans le modele par une MuteVar.
  @param _value La value de la borne suprieure
  @see Model::SetMuteVarUB
 */
void NumVar::SetUpperBound(float _value)
{
	if(!modele)
    cerr << "Warning in NumVar::SetUpperBound : undefined model" << endl;
	if(modele) modele->SetMuteVarUB(varId,_value);
}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter permettant de dfinir si la borne infrieure de la variable est stricte ou non
  @return true si la borne infrieure est stricte
  @see Model::GetMuteVar   MuteVar::IsLowerBoundStrict
*/
bool NumVar::IsLowerBoundStrict() const
{
	if(!modele)
    cerr << "Warning in NumVar::IsLowerBoundStrict : undefined model" << endl;
  return (modele?modele->GetMuteVar(varId).IsLowerBoundStrict():false);
}

/**
  Geter permettant de dfinir si la borne infrieure de la variable est stricte ou non
  @param _value= true si la borne infrieure doit tre stricte
  @see Model::SetMuteVarLBStrict
 */
void NumVar::SetLowerBoundStrict( bool _value )
{
	if(!modele)
    cerr << "Warning in NumVar::SetLowerBoundStrict : undefined model" << endl;
	if(modele) modele->SetMuteVarLBStrict(varId,_value);
}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter permettant de dfinir si la borne suprieure de la variable est stricte ou non
  @return true si la borne suprieure est stricte
  @see Model::GetMuteVar   MuteVar::IsUpperBoundStrict
*/
bool NumVar::IsUpperBoundStrict() const
{
	if(!modele)
    cerr << "Warning in NumVar::IsUpperBoundStrict : undefined model" << endl;
  return (modele?modele->GetMuteVar(varId).IsUpperBoundStrict():false);
}

/**
  Geter permettant de dfinir si la borne suprieure de la variable est stricte ou non
  @param _value= true si la borne suprieure doit tre stricte
  @see Model::SetMuteVarUBStrict
 */
void NumVar::SetUpperBoundStrict( bool _value )
{
	if(!modele)
    cerr << "Warning in NumVar::SetUpperBoundStrict : undefined model" << endl;
	if(modele) modele->SetMuteVarUBStrict(varId,_value);
}

/*-----------------------------------------------------------------------------------------------------------*/
/**
  Geter du type de la variable
  @return Le type de la variable
  @see Model::GetMuteVar MuteVar::GetType
 */

MuteVar::VarType NumVar::GetType() const
{
	if(!modele)
    cerr << "Warning in NumVar::GetType : undefined model" << endl;
  return (modele?modele->GetMuteVar(varId).GetType():MuteVar::FLOAT);
}

/**
  Seter du type de la variable
  @param Le type que doit prendre la variable
  @see Model::SetType
 */
void NumVar::SetType( MuteVar::VarType _type)
{
	if(!modele)
    cerr << "Warning in NumVar::SetType : undefined model" << endl;
	if(modele) modele->SetType(varId,_type);
}


#ifdef COMPILE_FOR_PYTHON
       #include "py_NumVar.cxx"
#endif


}
