/*
	Copyright (c) 2005, Quentin Lequy and Romain Gaucher
	All rights reserved.

	Redistribution and use in source and binary forms, with or 
	without modification, are permitted provided that the following 
	conditions are met:

		* Redistributions of source code must retain the above copyright 
		  notice, this list of conditions and the following disclaimer.

		* Redistributions in binary form must reproduce the above copyright 
		  notice, this list of conditions and the following disclaimer in the 
		  documentation and/or other materials provided with the distribution.

		* Neither the name of Quentin Lequy or Romain Gaucher nor the names 
		  of its contributors may be used to endorse or promote products 
		  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __CONSTRAINTBUILDER_H
#define __CONSTRAINTBUILDER_H

#include "Expr.h"
#include <list>

namespace Modelib {

class ostream;


///Classe servant  les Contraintes lineaire et les bornes
/**
  Elle est constitu:
   - d'une liste d'expressions linaire
   - d'une liste d'operateurs parmi l'numeration Operateur
   - d'un boolen permettant de savoir si la contrainte est une borne
  Elle peut representer des contraintes enchaine: 0<=3x+4<=7u+3y<4y-5x<18
  Elle est construite soit  partir de deux expressions, de deux constraintes ou d'une contraintes et d'une expression
  Elle est directement exploitables par la classe Model par le biais de la dclaration en friend de Model.
  Les raisons sont les mme que pour la classe Expr,  savoir que ConstraintBuilder n'a volontairement pas de geters,
  de seters ou d'iterateurs sur les listes d'Expr ou d'Operateurs qui sont prives.
 */

class ConstraintBuilder
{
  public:
    /**Oprateurs de comparaison*/
    enum Operateur {EGAL=0,DIFFERENT,INFERIEUR,SUPERIEUR,INFSTRICT,SUPSTRICT};

  protected:
    std::list<Expr> expressions;       //!< liste d'expressions
    std::list<Operateur> operateurs;   //!< liste d'operateurs
    bool isBound;                      //!< boolen permettant de savoir si l'expressions est une borne
    
    bool CheckIfIsBound();
    void Add(Operateur,const std::list<Expr> &,const std::list<Operateur> &);
   
    friend class Model;
    friend std::ostream & operator<< (std::ostream & flux, const ConstraintBuilder & c);

  public:
    /**Constructeur par dfaut*/
    ConstraintBuilder(): isBound(false) {}
    ConstraintBuilder( const Expr & , ConstraintBuilder::Operateur , const Expr &);
    /**Constructeur par recopie*/
    ConstraintBuilder(const ConstraintBuilder & c) { *this = c; }
    /**Oprateur de copie*/
    ConstraintBuilder & operator= ( const ConstraintBuilder & _c)
    {
        if( this != &_c )   //permet d'economiser la recopie de liste
        {
          isBound = _c.isBound;
          expressions = _c.expressions;
          operateurs = _c.operateurs;
        }
        return *this;
    }
    /**Permet de savoir si la contrainte est une borne*/
    inline bool IsBound() const { return isBound;}
  
    //operateurs permettant d'enchainer les contraintes
    ConstraintBuilder & operator== ( const ConstraintBuilder & );
    ConstraintBuilder & operator!= ( const ConstraintBuilder & );
    ConstraintBuilder & operator<= ( const ConstraintBuilder & );
    ConstraintBuilder & operator>= ( const ConstraintBuilder & );
    ConstraintBuilder & operator< ( const ConstraintBuilder & );
    ConstraintBuilder & operator> ( const ConstraintBuilder & );
  
    //operateur permettant de terminer un enchainement
    ConstraintBuilder & operator== ( const Expr & );
    ConstraintBuilder & operator!= ( const Expr & );
    ConstraintBuilder & operator<= ( const Expr & );
    ConstraintBuilder & operator>= ( const Expr & );
    ConstraintBuilder & operator< ( const Expr & );
    ConstraintBuilder & operator> ( const Expr & );

    ~ConstraintBuilder() {}
};

//operateurs permettant de crer une expression a partir de deux expressions
ConstraintBuilder operator== ( const Expr &, const Expr & );
ConstraintBuilder operator!= ( const Expr &, const Expr & );
ConstraintBuilder operator<= ( const Expr &, const Expr & );
ConstraintBuilder operator>= ( const Expr &, const Expr & );
ConstraintBuilder operator< ( const Expr &, const Expr & );
ConstraintBuilder operator> ( const Expr &, const Expr & );

std::ostream & operator<< (std::ostream & flux,const ConstraintBuilder & c);

}

#endif
